from graphql.type.scalars import GraphQLScalarType

from .type_converter import REGISTRY


def _process_scalar(cls, *, name, description, serialize, parse_value, parse_literal):
    if name is None:
        name = cls.__name__

    REGISTRY[cls] = GraphQLScalarType(
        name=name,
        description=description,
        serialize=serialize,
        parse_value=parse_value,
        parse_literal=parse_literal,
    )

    return cls


def identity(x):
    return x


def scalar(
    cls=None,
    *,
    name=None,
    description=None,
    serialize=identity,
    parse_value=None,
    parse_literal=None
):
    """Annotates a class or type as a GraphQL custom scalar.

    Example usages:

    >>> strawberry.scalar(
    >>>     datetime.date,
    >>>     serialize=lambda value: value.isoformat(),
    >>>     parse_value=datetime.parse_date
    >>> )

    >>> Base64Encoded = strawberry.scalar(
    >>>     NewType("Base64Encoded", bytes),
    >>>     serialize=base64.b64encode,
    >>>     parse_value=base64.b64decode
    >>> )

    >>> @strawberry.scalar(
    >>>     serialize=lambda value: ",".join(value.items),
    >>>     parse_value=lambda value: CustomList(value.split(","))
    >>> )
    >>> class CustomList:
    >>>     def __init__(self, items):
    >>>         self.items = items

    """

    if parse_value is None:
        parse_value = cls

    def wrap(cls):
        return _process_scalar(
            cls,
            name=name,
            description=description,
            serialize=serialize,
            parse_value=parse_value,
            parse_literal=parse_literal,
        )

    if cls is None:
        return wrap

    return wrap(cls)



from os.path import abspath, dirname, join


def render_graphiql_page():
    dir_path = abspath(join(dirname(__file__), ".."))
    graphiql_html_file = f"{dir_path}/static/graphiql.html"

    html_string = None
    with open(graphiql_html_file, "r") as f:
        html_string = f.read()
    return html_string



import json

from flask import Response, abort, render_template_string, request
from flask.views import View
from graphql import graphql_sync
from graphql.error import format_error as format_graphql_error
from graphql.type.schema import GraphQLSchema

from .graphiql import render_graphiql_page


class GraphQLView(View):
    schema = None
    graphiql = True

    methods = ["GET", "POST"]

    def __init__(self, schema, graphiql=True):
        self.schema = schema
        self.graphiql = graphiql

        if not self.schema:
            raise ValueError("You must pass in a schema to GraphQLView")

        if not isinstance(self.schema, GraphQLSchema):
            raise ValueError("A valid schema is required to be provided to GraphQLView")

    def render_template(self, request, template=None):
        return render_template_string(template)

    def dispatch_request(self):
        if "text/html" in request.environ.get("HTTP_ACCEPT", ""):
            if not self.graphiql:
                abort(404)

            template = render_graphiql_page()
            return self.render_template(request, template=template)

        data = request.json

        try:
            query = data["query"]
            variables = data.get("variables")
            operation_name = data.get("operationName")

        except KeyError:
            return Response("No valid query was provided for the request", 400)

        context = dict(request=request)
        result = graphql_sync(
            self.schema,
            query,
            variable_values=variables,
            context_value=context,
            operation_name=operation_name,
        )

        response_data = {"data": result.data}

        if result.errors:
            response_data["errors"] = [
                format_graphql_error(err) for err in result.errors
            ]

        return Response(
            json.dumps(response_data),
            status=400 if result.errors else 200,
            content_type="application/json",
        )



import collections
import typing


def is_list(annotation):
    """Returns True if annotation is a typing.List"""

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == list


def is_union(annotation):
    """Returns True if annotation is a typing.Union"""

    annotation_origin = getattr(annotation, "__origin__", None)

    return annotation_origin == typing.Union


def is_optional(annotation):
    """Returns True if the annotation is typing.Optional[SomeType]"""

    # Optionals are represented as unions

    if not is_union(annotation):
        return False

    types = annotation.__args__

    # A Union to be optional needs to have at least one None type
    return any([x == None.__class__ for x in types])  # noqa:E711


def get_optional_annotation(annotation):
    types = annotation.__args__
    non_none_types = [x for x in types if x != None.__class__]  # noqa:E711

    return non_none_types[0]


def get_list_annotation(annotation):
    return annotation.__args__[0]


def is_generic(annotation) -> bool:
    """Returns True if the annotation is or extends a generic."""
    return (
        isinstance(annotation, type)
        and issubclass(annotation, typing.Generic)  # type:ignore
        or isinstance(annotation, typing._GenericAlias)  # type:ignore
        and annotation.__origin__
        not in (
            list,
            typing.Union,
            tuple,
            typing.ClassVar,
            collections.abc.AsyncGenerator,
        )
    )


def is_type_var(annotation) -> bool:
    """Returns True if the annotation is a TypeVar."""

    return isinstance(annotation, typing.TypeVar)  # type:ignore


def has_type_var(annotation) -> bool:
    """
    Returns True if the annotation or any of
    its argument have a TypeVar as argument.
    """
    return any(
        is_type_var(arg) or has_type_var(arg)
        for arg in getattr(annotation, "__args__", [])
    )


def get_actual_type(annotation, types_replacement_map):
    """Returns a copy of an annotation by replacing TypeVar"""
    if is_type_var(annotation):
        return types_replacement_map[annotation.__name__]

    if has_type_var(annotation):
        return annotation.copy_with(
            tuple(
                get_actual_type(arg, types_replacement_map)
                for arg in annotation.__args__
            )
        )

    return annotation



import dataclasses
from collections.abc import AsyncGenerator

from graphql import (
    GraphQLBoolean,
    GraphQLFloat,
    GraphQLID,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLString,
)

from .exceptions import MissingTypesForGenericError
from .scalars import ID
from .union import union
from .utils.str_converters import capitalize_first, to_camel_case
from .utils.typing import is_generic, is_union


REGISTRY = {
    str: GraphQLString,
    int: GraphQLInt,
    float: GraphQLFloat,
    bool: GraphQLBoolean,
    ID: GraphQLID,
}


def copy_annotation_with_types(annotation, *types):
    # unfortunately we need to have this back and forth with GraphQL types
    # and original classes for the time being. In future we might migrate
    # away from GraphQL-Core or build our own tiny class wrapper that will
    # help with generic types, but this works ok for now.
    origin = annotation.__origin__
    graphql_type = origin.graphql_type
    TypeClass = type(graphql_type)

    types_replacement_map = dict(
        zip([param.__name__ for param in origin.__parameters__], types)
    )
    copied_name = (
        "".join([capitalize_first(type.__name__) for type in types]) + graphql_type.name
    )

    extra_kwargs = {
        "description": graphql_type.description,
        "interfaces": graphql_type._interfaces,
    }

    def get_fields():
        origin_fields = dict(
            (to_camel_case(f.name), f) for f in dataclasses.fields(origin)
        )
        fields = graphql_type._fields(types_replacement_map)

        for field_name in fields.keys():
            origin_field = origin_fields[field_name]

            if is_generic(origin_field.type):
                fields[field_name] = copy_annotation_with_types(
                    origin_field.type, *types
                )

        return fields

    new_type = TypeClass(copied_name, get_fields, **extra_kwargs)

    # we use _copied to easily find a previously created type,
    # this is used by unions when needing to find the proper GraphQL
    # type when returning something
    # TODO: we could use this to prevent running all the code above

    if not hasattr(origin, "_copies"):
        origin._copies = {}

    origin._copies[types] = new_type

    return new_type


# TODO: make so that we don't pass force optional
# we use that when trying to get the type for a
# option field (which can either be a scalar or an object type)
def get_graphql_type_for_annotation(
    annotation, field_name: str, force_optional: bool = False
):
    # TODO: this might lead to issues with types that have a field value
    is_field_optional = force_optional

    if is_generic(annotation):
        types = getattr(annotation, "__args__", None)

        if types is None:
            raise MissingTypesForGenericError(field_name, annotation)

        graphql_type = copy_annotation_with_types(annotation, *types)
    elif hasattr(annotation, "graphql_type"):
        graphql_type = annotation.graphql_type
    else:
        annotation_name = getattr(annotation, "_name", None)

        if annotation_name == "List":
            list_of_type = get_graphql_type_for_annotation(
                annotation.__args__[0], field_name
            )

            list_type = GraphQLList(list_of_type)

            return list_type if is_field_optional else GraphQLNonNull(list_type)

        annotation_origin = getattr(annotation, "__origin__", None)

        if annotation_origin == AsyncGenerator:
            # async generators are used in subscription, we only need the yield type
            # https://docs.python.org/3/library/typing.html#typing.AsyncGenerator
            return get_graphql_type_for_annotation(annotation.__args__[0], field_name)

        elif is_union(annotation):
            types = annotation.__args__
            non_none_types = [x for x in types if x != None.__class__]  # noqa:E711

            # optionals are represented as Union[type, None]

            if len(non_none_types) == 1:
                is_field_optional = True
                graphql_type = get_graphql_type_for_annotation(
                    non_none_types[0], field_name, force_optional=True
                )
            else:
                is_field_optional = None.__class__ in types

                graphql_type = union(field_name, types).graphql_type
        else:
            graphql_type = REGISTRY.get(annotation)

    if not graphql_type:
        raise ValueError(f"Unable to get GraphQL type for {annotation}")

    if is_field_optional:
        return graphql_type

    return GraphQLNonNull(graphql_type)



import enum
import typing
from dataclasses import is_dataclass
from datetime import date, datetime, time

from ..exceptions import UnsupportedTypeError
from .str_converters import to_camel_case
from .typing import get_list_annotation, get_optional_annotation, is_list, is_optional


SCALAR_TYPES = [int, str, float, bytes, bool, datetime, date, time]


class _Unset:
    def __str__(self):
        return ""

    def __bool__(self):
        return False


UNSET = _Unset()


def is_unset(value: typing.Any) -> bool:
    return value is UNSET


def convert_args(
    value: typing.Union[typing.Dict[str, typing.Any], typing.Any],
    annotation: typing.Union[typing.Dict[str, typing.Type], typing.Type],
):
    """Converts a nested dictionary to a dictionary of actual types.

    It deals with conversion of input types to proper dataclasses and
    also uses a sentinel value for unset values."""

    if annotation == {}:
        return value

    if value is None:
        return None

    if is_unset(value):
        return value

    if is_optional(annotation):
        annotation = get_optional_annotation(annotation)

    # TODO: change this to be a is_scalar util and make sure it works with any scalar
    if getattr(annotation, "__supertype__", annotation) in SCALAR_TYPES:
        return value

    # Convert Enum fields to instances using the value. This is safe
    # because graphql-core has already validated the input.
    if isinstance(annotation, enum.EnumMeta):
        return annotation(value)  # type: ignore

    if is_list(annotation):
        annotation = get_list_annotation(annotation)

        return [convert_args(x, annotation) for x in value]

    fields = None

    # we receive dicts when converting resolvers arguments to
    # actual types
    if isinstance(annotation, dict):
        fields = annotation.items()

    elif is_dataclass(annotation):
        fields = annotation.__dataclass_fields__.items()

    if fields:
        kwargs = {}

        for name, field in fields:
            dict_name = name

            if hasattr(field, "field_name") and field.field_name:
                dict_name = field.field_name
            else:
                dict_name = to_camel_case(name)

            # dataclasses field have a .type attribute
            if hasattr(field, "type"):
                field_type = field.type
            # meanwhile when using dicts the value of the field is
            # the actual type, for example in: { 'name': str }
            else:
                field_type = field

            kwargs[name] = convert_args(value.get(dict_name, UNSET), field_type)

        if is_dataclass(annotation):
            return annotation(**kwargs)  # type: ignore

        return kwargs

    raise UnsupportedTypeError(annotation)
